"""
Functions to create and process rotation and transformation matrices.
Some conversions to and from Euler angles.
"""
import numpy as np
from .quaternion import Quaternion


def rot_x(alfa):
    return np.array(
        [[1, 0, 0], [0, np.cos(alfa), -np.sin(alfa)], [0, np.sin(alfa), np.cos(alfa)]]
    )


def rot_y(alfa):
    return np.array(
        [[np.cos(alfa), 0, np.sin(alfa)], [0, 1, 0], [-np.sin(alfa), 0, np.cos(alfa)]]
    )


def rot_z(alfa):
    return np.array(
        [[np.cos(alfa), -np.sin(alfa), 0], [np.sin(alfa), np.cos(alfa), 0], [0, 0, 1]]
    )


def translation(x, y, z):
    return np.array(
        [[1, 0, 0, x], [0, 1, 0, y], [0, 0, 1, z], [0, 0, 0, 1]], dtype="float64"
    )


def pose_x(alfa, x, y, z):
    """ Homogenous transform with rotation around x-axis and translation. """
    return np.array(
        [
            [1, 0, 0, x],
            [0, np.cos(alfa), -np.sin(alfa), y],
            [0, np.sin(alfa), np.cos(alfa), z],
            [0, 0, 0, 1],
        ]
    )


def pose_y(alfa, x, y, z):
    """ Homogenous transform with rotation around y-axis and translation. """
    return np.array(
        [
            [np.cos(alfa), 0, np.sin(alfa), x],
            [0, 1, 0, y],
            [-np.sin(alfa), 0, np.cos(alfa), z],
            [0, 0, 0, 1],
        ]
    )


def pose_z(alfa, x, y, z):
    """ Homogenous transform with rotation around z-axis and translation. """
    return np.array(
        [
            [np.cos(alfa), -np.sin(alfa), 0, x],
            [np.sin(alfa), np.cos(alfa), 0, y],
            [0, 0, 1, z],
            [0, 0, 0, 1],
        ]
    )


def quat_distance(qa: Quaternion, qb: Quaternion):
    """ Half of the rotation angle to bring qa to qb."""
    return np.arccos(np.abs(qa.elements @ qb.elements))


def tf_inverse(T):
    """ Efficient inverse of a homogenous transform.
    (Normal matrix inversion would be a bad idea.)
    Returns a copy, not inplace!
    """
    Ti = np.eye(4)
    Ti[:3, :3] = T[:3, :3].transpose()
    Ti[:3, 3] = np.dot(-Ti[:3, :3], T[:3, 3])
    return Ti


def rpy_to_rot_mat(rxyz):
    """ Convert roll, pitch and yaw angles to a rotation matrix."""
    r_x, r_y, r_z = rxyz[0], rxyz[1], rxyz[2]
    return rot_x(r_x) @ rot_y(r_y) @ rot_z(r_z)


def rotation_matrix_to_rpy(R):
    """
    Convert Roll, pitch and yaw angles to a rotation matrix
    (intrinsic XYZ Euler angles).

    I'm not sure if this is the best implementation.
    Look into the one scipy provides.
    """
    r11, r12, r13 = R[0]
    _, _, r23 = R[1]
    _, _, r33 = R[2]

    r_x = np.arctan2(-r23, r33)
    r_y = np.arctan2(r13, np.sqrt(r11 ** 2 + r12 ** 2))
    r_z = np.arctan2(-r12, r11)

    return [r_x, r_y, r_z]




import numpy as np
from math import sqrt

DTYPE = np.double


def norm_l1(double[:,:] D1, double[:, :] D2):
    
    cdef Py_ssize_t N1 = D1.shape[0]
    cdef Py_ssize_t N2 = D2.shape[0]
    cdef Py_ssize_t K = D2.shape[1]
        
    result = np.zeros((N1, N2), dtype=DTYPE)
    cdef double[:, :] result_view = result
    
    cdef Py_ssize_t i, j, k
    cdef double tmp
    for i in range(N1):
        for j in range(N2):
            tmp = 0
            for k in range(K):
                tmp += abs(D1[i, k] - D2[j, k])
            result_view[i, j] = tmp
    
    return result

def norm_l2(double[:,:] D1, double[:, :] D2):
    
    cdef Py_ssize_t N1 = D1.shape[0]
    cdef Py_ssize_t N2 = D2.shape[0]
    cdef Py_ssize_t K = D2.shape[1]
        
    result = np.zeros((N1, N2), dtype=DTYPE)
    cdef double[:, :] result_view = result
    
    cdef Py_ssize_t i, j, k
    cdef double tmp
    for i in range(N1):
        for j in range(N2):
            tmp = 0
            for k in range(K):
                tmp += (D1[i, k] - D2[j, k]) ** 2
            result_view[i, j] = sqrt(tmp)
    
    return result

def sum_squared(double[:,:] D1, double[:, :] D2):
    
    cdef Py_ssize_t N1 = D1.shape[0]
    cdef Py_ssize_t N2 = D2.shape[0]
    cdef Py_ssize_t K = D2.shape[1]
        
    result = np.zeros((N1, N2), dtype=DTYPE)
    cdef double[:, :] result_view = result
    
    cdef Py_ssize_t i, j, k
    cdef double tmp
    for i in range(N1):
        for j in range(N2):
            tmp = 0
            for k in range(K):
                tmp += (D1[i, k] - D2[j, k]) ** 2
            result_view[i, j] = tmp
    
    return result

def weighted_sum_squared(double[:,:] D1, double[:, :] D2, double[:] weights):
    
    cdef Py_ssize_t N1 = D1.shape[0]
    cdef Py_ssize_t N2 = D2.shape[0]
    cdef Py_ssize_t K = D2.shape[1]
        
    result = np.zeros((N1, N2), dtype=DTYPE)
    cdef double[:, :] result_view = result
    
    cdef Py_ssize_t i, j, k
    cdef double tmp
    for i in range(N1):
        for j in range(N2):
            tmp = 0
            for k in range(K):
                tmp += weights[k] * (D1[i, k] - D2[j, k]) ** 2
            result_view[i, j] = tmp
    
    return result

def norm_infinity(double[:,:] D1, double[:, :] D2):
    
    cdef Py_ssize_t N1 = D1.shape[0]
    cdef Py_ssize_t N2 = D2.shape[0]
    cdef Py_ssize_t K = D2.shape[1]
        
    result = np.zeros((N1, N2), dtype=DTYPE)
    cdef double[:, :] result_view = result
    
    cdef Py_ssize_t i, j, k
    cdef double tmp
    for i in range(N1):
        for j in range(N2):
            tmp = 0
            for k in range(K):
                if abs((D1[i, k] - D2[j, k])) > tmp:
                    tmp = abs(D1[i, k] - D2[j, k])
            result_view[i, j] = tmp
    
    return result


"""
A small extension of the pyquaternion library.
"""
import numpy as np
import pyquaternion


class Quaternion(pyquaternion.Quaternion):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def random_near(self, distance):
        """ Generate random unit quaternion near a given quaternion.
      Based on the 'SO3StateSpace' implementation in the open motion
      planning library (http://ompl.kavrakilab.org/).
      Params:
          other: a Quaternion around wich to sample
          distance: the maximum distance from the given Quaternion
      """

        if distance > 0.25 * np.pi:
            return super().random()

        d = np.random.uniform()
        g1, g2, g3 = np.random.normal(size=(3,))
        q_delta = Quaternion(axis=[g1, g2, g3], angle=(2 * d ** (1 / 3) * distance))

        return self * q_delta  # multiplication returns a new quaterion



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa: F401 unused import

def get_default_axes3d(xlim=[-1, 1], ylim=[-1, 1], zlim=[-1, 1]):
    """ Create a default `mpl_toolkits.mplot3d.Axes3D` object with default
    axis limits on all axis from -1 to 1, and labels on the axes.
    """
    fig = plt.figure()
    ax = fig.gca(projection="3d")
    ax.set_xlim3d(xlim)
    ax.set_ylim3d(ylim)
    ax.set_zlim3d(zlim)
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")
    return fig, ax


def plot_reference_frame(ax, tf=None, arrow_length=0.2):
    """ Plot xyz-axes on axes3d object
    Parameters
    ----------
    ax : mpl_toolkits.mplot3d.Axes3D
        Axes object for 3D plotting.
    tf : np.array of float
        Transform to specify location of axes. Plots in origin if None.
    l : float
        The length of the axes plotted.
    """
    l = arrow_length
    x_axis = np.array([[0, l], [0, 0], [0, 0]])
    y_axis = np.array([[0, 0], [0, l], [0, 0]])
    z_axis = np.array([[0, 0], [0, 0], [0, l]])

    if tf is not None:
        # rotation
        x_axis = np.dot(tf[:3, :3], x_axis)
        y_axis = np.dot(tf[:3, :3], y_axis)
        z_axis = np.dot(tf[:3, :3], z_axis)
        # translation [:, None] numpian way to change shape (add axis)
        x_axis = x_axis + tf[:3, 3][:, None]
        y_axis = y_axis + tf[:3, 3][:, None]
        z_axis = z_axis + tf[:3, 3][:, None]

    ax.plot(x_axis[0], x_axis[1], x_axis[2], "-", c="r")
    ax.plot(y_axis[0], y_axis[1], y_axis[2], "-", c="g")
    ax.plot(z_axis[0], z_axis[1], z_axis[2], "-", c="b")



import numpy as np


def apply_cost_function(data, fun):
    """ Calculate tranitions costs in between states. """
    res = []
    for i in range(1, len(data)):
        res.append(fun(data[i - 1], data[i]))
    return res


def calculate_value_function(transition_costs):
    """Recursively apply the bellman equation from the end to the start. """
    state_dim = [tc.shape[0] for tc in transition_costs]
    state_dim.append(transition_costs[-1].shape[1])

    V = [np.zeros(d) for d in state_dim]
    V_ind = [np.zeros(d) for d in state_dim]

    for i in range(len(state_dim) - 2, -1, -1):
        rhs = transition_costs[i] + V[i + 1]
        V[i] = np.min(rhs, axis=1)
        V_ind[i] = np.argmin(rhs, axis=1)

    return V_ind, V


def calculate_value_function_with_state_cost(transition_costs, state_costs):
    """Recursively apply the bellman equation from the end to the start. """
    assert len(transition_costs) == (len(state_costs) - 1)

    state_dim = [tc.shape[0] for tc in transition_costs]
    state_dim.append(transition_costs[-1].shape[1])

    V = np.copy(state_costs)  #  intial value function just the state cost?
    V_ind = [np.zeros(d) for d in state_dim]

    for i in range(len(state_dim) - 2, -1, -1):
        rhs = transition_costs[i] + V[i + 1]
        V[i] = V[i] + np.min(rhs, axis=1)
        V_ind[i] = np.argmin(rhs, axis=1)

    return V_ind, V


def extract_shortest_path(data, V_ind, V):
    """ Start in stage one with the state that has to lowest value in V
    and follow the index trace from then on.
    """
    res = []
    i_start = np.argmin(V[0])
    i_next = i_start
    for i_stage in range(len(data)):
        res.append(data[i_stage][i_next])
        i_next = V_ind[i_stage][i_next]

    return res


def shortest_path(Q, cost_function):
    costs = apply_cost_function(Q, cost_function)
    actions, values = calculate_value_function(costs)

    path = extract_shortest_path(Q, actions, values)
    path_length = np.min(values[0])
    return {"success": True, "path": path, "length": path_length}


def shortest_path_with_state_cost(states, state_costs, cost_function):
    costs = apply_cost_function(states, cost_function)
    actions, values = calculate_value_function_with_state_cost(costs, state_costs)

    path = extract_shortest_path(states, actions, values)
    path_length = np.min(values[0])
    return {"success": True, "path": path, "length": path_length}




from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize

extensions = [Extension("acrolib.cost_functions", ["src/acrolib/cost_functions.pyx"])]

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="acrolib",
    version="0.0.7",
    author="Jeroen De Maeyer",
    author_email="jeroen.demaeyer@kuleuven.be",
    description="General utilities for my robotics research at ACRO.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JeroenDM/acrolib",
    packages=find_packages(where="src"),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    ext_modules=cythonize(extensions),
    python_requires=">=3.6",
)



import numpy as np
from numpy.random import uniform
from enum import Enum


class SampleMethod(Enum):
    random_uniform = 0
    deterministic_uniform = 1


class Sampler:
    def __init__(self):
        self.halton_sampler: HaltonSampler = None

    def sample(self, num_samples: int, sample_dim: int, method: SampleMethod):
        if method == SampleMethod.random_uniform:
            return np.random.rand(num_samples, sample_dim)
        elif method == SampleMethod.deterministic_uniform:
            if self.halton_sampler is None:
                self.halton_sampler = HaltonSampler(sample_dim)
            # make sure an existing sampler has the correct dimension
            assert self.halton_sampler.dim == sample_dim
            return self.halton_sampler.get_samples(num_samples)
        else:
            raise NotImplementedError(f"Unkown sampling method: {method}")


def vdc(n, base=2):
    """ Create van der Corput sequence

    source for van der Corput and Halton sampling code
    https://laszukdawid.com/2017/02/04/halton-sequence-in-python/
    """
    vdc, denom = 0, 1
    while n:
        denom *= base
        n, remainder = divmod(n, base)
        vdc += remainder / denom
    return vdc


def next_prime():
    def is_prime(num):
        "Checks if num is a prime value"
        for i in range(2, int(num ** 0.5) + 1):
            if (num % i) == 0:
                return False
        return True

    prime = 3
    while 1:
        if is_prime(prime):
            yield prime
        prime += 2


class HaltonSampler:
    def __init__(self, dim):
        self.dim = dim

        # setup primes for every dimension
        prime_factory = next_prime()
        self.primes = []
        for _ in range(dim):
            self.primes.append(next(prime_factory))

        # init counter for van der Corput sampling
        self.cnt = 1

    def get_samples(self, n):
        seq = []
        for d in range(self.dim):
            base = self.primes[d]
            seq.append([vdc(i, base) for i in range(self.cnt, self.cnt + n)])
        self.cnt += n
        return np.array(seq).T
